#!/usr/bin/perl

eval "exec /usr/local/bin/perl -s $0"
  if $running_under_some_shell;

# @(#)@ dorequest	1.19 - dorequest.pl
#
# This program handles the enqueued requests for the mail server.
#
# Arguments: address file [ encoding [ size [ list ] ] ]
#
#   address : where to send the information to.
#	      If left empty, no splitting is done, and the result
#	      is written to stdout.
#
#   file    : the file to send.
#
#   encoding: how to encode it: uuencode, btoa or plain.
#             Default is btoa.
#
#   limit   : how many bytes per transmission.
#             Default is 32768
#
#   parts   : comma- or blank-separated list of part numbers.
#             When used, only these parts are sent.
#
################ debugging section ################
#
# set these variables using perl -s
#
$debug = 0 unless $debug;
$nolog = 0 unless $nolog;
$nosync = 0 unless $nosync;
#
################ configuration section ################
#
# where to find these
#
$btoa = "/usr/local/bin/btoa";
# Note uue is an enhanced version of uuencode
$uuencode = "/usr/local/bin/uue";
#$uuencode = "/usr/local/bin/uuencode";
$uux = "/usr/bin/uux";
#
# where to log transactions (if file exists and writable).
# set to "" to disable logging.
#
$logfile = "/usr/lib/mserv/.logfile";
#
do "/usr/lib/mserv/rfc822.pl";
#
# exclusive locking of multiple jobs
$lockfile = "/usr/lib/mserv/.lockfile";
#
# there are probably more system dependent things in sub synchronize and 
# in sub xfer.
#
$domain = "mh.nl";
$host = `/bin/hostname`; chop ($host);
$timezone = "MET";
$smail_lib = "/usr/lib/smail";
$smail_log = $smail_lib . "/.logfile";
$smail_router = $smail_lib . "/smail -A -R";
$sendmail = $smail_lib . "/sendmail-t";
#
$sender = "From: Multihouse Mail Server <MAILER-DAEMON@$domain>";
#
# optional mail headers. set to "" if not wanted.
#
$x_server = "X-Server: Multihouse Research Mail Server Backend 1.19";
$x_info = "X-Info: send mail to <postmaster@mh.nl>";
#
################ argument processing ################

# note: the "parts" argument is omitted from the usage message
if ( $#ARGV < 1 | $#ARGV > 4 ) {
  print stderr "usage: dorequest address file [ encoding [ size ] ]\n";
  print stderr "call was: $0 \"", join ("\" \"", @ARGV), "\"\n";
  exit 1;
}

# address to send to
$address = shift (@ARGV);

# the requested file
$file = shift (@ARGV);

# encoding to be used [optional]
# note that the case handling on $encoding defaults to "btoa" if
# it is not "uuencode" or "plain"
$encoding = ($#ARGV >= 0) ? shift (@ARGV) : "btoa";

# byte limit on transmissions [optional]
$limit = ($#ARGV >= 0) ? shift (@ARGV) : 32*1024;
$limit = 1024 if $limit < 1024;
$limit = 256*1024 if $limit > 256*1024;

# list of parts [optional]
$plist = ($#ARGV >= 0) ? " " . join (" ", split (/[ \t,]/, shift (@ARGV))) . " " : "";

if ( $address eq "" ) {
  $limit = 0;
  $nolog = 1;
  $nosync = 1;
}

################ end of argument processing ################

# cut down system load

do synchronise ();

# Proceed

$tmpfile_prefix = "/tmp/drq$$.";

# get last part (basename) of the requested file
@tmp = split (/\//, $file); 
$fname = $tmp[$#tmp];

# prepare the command to use
if ( $encoding eq "btoa" ) {

  $code = "btoa encoded";
  # btoa has lines of 78 chars
  $lines = int ($limit / 79) - 2;
  $cmd = "$btoa < $file";
  $ct = "B";
}
elsif ( $encoding eq "plain" ) {

  $code = "ascii";
  $cmd = "cat < $file";
  $ct = "A";
  $lines = 0;
}
else {

  $code = "uuencoded";
  # uuencode has lines of 61 chars
  $lines = int ($limit / 62) - 2;
  $cmd = "$uuencode \"$fname\" < $file";
  $ct = "U";
}

if ( $lines > 0 && $limit > 0 ) {
  # encode and split
  system $cmd . " | split -$lines - $tmpfile_prefix";
  # now gather all the parts, and send them
  $wild = $tmpfile_prefix . "??";
  @files = sort <${wild}>; 
  $count = $#files + 1;

  if ( $debug ) {
    print stderr "sending ", $count, " files: ", join(" ",@files), "\n";
  }
}
else {
  system $cmd . " > $tmpfile_prefix.aa";
  @files = "$tmpfile_prefix.aa";
  $count = 1;
}

$num = 1;
while ( $#files >= 0 ) {

  $the_file = shift (@files);

  if ( $plist ne "" && index ($plist, " $num ") < 0 ) {
    # use "next" to activate "continue" block
    next;
  }

  # form "part xx of yy" message
  $part = ( $count == 1 ) ? "complete" : "part $num of $count";

  # send it
  if ( open (part, $the_file) ) {
    if ( $address eq "" ) {
      do copy ("STDOUT");
    }
    else {
      do xfer ();
    }
    close (part);
  }

  # write a log message
  # $xfer_size is set by 'xfer()'
  do writelog ("\"$address\" $file $ct$num/$count $xfer_size")
    if $address ne "";

} continue {
  # remove this part, and advance
  unlink ($the_file);
  $num++;
}

# remove the lock, if any
unlink $lockfile if $lockfile;

# that's it ...
exit 0;

################ subroutines ################

sub headers {
  local ($mailer, $all) = @_;
  local (@days) = ("Sun","Mon","Tue","Wed","Thu","Fri","Sat");
  local (@months) = ("Jan","Feb","Mar","Apr","May","Jun",
                     "Jul","Aug","Sep","Oct","Nov","Dec");
  local ($size);

  # get time and tz difference
  $tim = time;
  @tm = gmtime ($tim);
  $tzdiff = -$tm[2];
  @tm = localtime ($tim);
  $tzdiff += $tm[2];
  $this_month = $months[$tm[4]];
  $this_day = $days[$tm[6]];

  $smail_date = sprintf ("%s %2d %02d:%02d", $this_month, $tm[3], $tm[2], $tm[1]);

  if ( $all ) {
    # primary headers
    $ln = sprintf ("From MAILER-DAEMON  %s %s %2d %2d:%02d:%02d 19%2d",
               $this_day, $this_month, $tm[3], 
	       $tm[2], $tm[1], $tm[0], $tm[5]) .
          " remote from $host\n" .
          "Received: by $domain (mserv1.19); " .
          sprintf ("%s, %d %2s %02d %2d:%02d:%02d %s (+%02d00)\n",
                   $this_day, $tm[3], $this_month, $tm[5], 
   	           $tm[2], $tm[1], $tm[0], $timezone,
	           ($tm[8] ? 1 : 0) + $tzdiff) .
          sprintf ("Message-Id: <$$.%02d%02d%02d%02d%02d@$domain>\n",
		   $tm[5], $tm[4], $tm[3], $tm[2], $tm[1]) .
	  "From: Multihouse Mail System <MAILER-DAEMON@$domain>\n";
    print $mailer $ln;
    $size += length ($ln);
  }
  else {
    print $mailer "$sender\n" if $sender;
    $size += length ($sender) + 1;
  }

  $ln = "To: $address\nSubject: $fname ($part) $code\n";
  if ( $all ) {
    $ln .= sprintf ("Date: %s, %d %2s %02d %2d:%02d:%02d %s\n",
                    $this_day, $tm[3], $this_month, $tm[5], 
   	            $tm[2], $tm[1], $tm[0], $timezone);
  }
  $ln .= "Precedence: bulk\n";
  $ln .= "$x_server\n" if $x_server;
  $ln .= "$x_info\n" if $x_info;
  print $mailer "$ln\n";
  $size += length ($ln) + 1;
}

sub copy {
  local ($mailer) = shift (@_);
  local ($size);
  local ($ln);

  $ln = "------ begin of $fname -- $code -- $part ------\n";
  $size = length ($ln);
  print $mailer $ln;
  while ( <part> ) {
    print $mailer $_;
    $size += length ($_);
  }
  $ln = "------ end of $fname -- $code -- $part ------\n";
  print $mailer $ln;
  $size += length ($ln);
}

sub xfer {
  # send the file via e-mail, avoid as much overhead as possible
  $did = 0;

  # parse the address
  do rfc822_parse_addresses ($address);

  # should be one, single destination
  if ( $#rfc822_addresses == 0 ) {

    $try = $rfc822_addresses[0];
    # call smail to (re)route
    $recipient = `$smail_router "$try"`;
    chop ($recipient);
    print stderr "real recipient: $recipient\n" if $debug;

    # take a look at the route. Note: smail will always deliver
    # a uucp path
    @try = split (/!/, $recipient);

    if ( $try[0] eq "hp4nl" ) {

      # it is our gateway ... use uux direcly
      $nexthop = shift (@try);
      $recipient = join ("!", @try);
      $cmd = "$uux - -r $nexthop!rmail '($recipient)'";
      if ( $debug ) {
        print stderr "[Would call \"$cmd\"]\n";
	do headers ("STDOUT", 1);
	$did = 1;
      }
      elsif ( open (mailer, "|$cmd") ) {
	# make the headers
	$xfer_size = do headers ("mailer", 1);
	# copy the file in
	$xfer_size += do copy ("mailer");

	# wrap-up
	close mailer;
	sleep (30) unless $num == $count;	# allow system to stabilize

	if ( -w $smail_log ) {
	  open (LOG, ">>" . $smail_log);
	  print LOG "$smail_date\t$xfer_size\tDaemon\t$cmd\n";
        }
	$did = 1;
      }
    }
  }

  # if we didn't send it - use regular method
  if ( !$did ) {

    # use "|/usr/lib/sendmail -t -odb" if you prefer sendmail

    if ( $debug ) {
      print stderr "[Would call \"$sendmail\"]\n";
      do headers ("STDOUT", 0);
    }
    elsif ( open (mailer, "|$sendmail") ) {
      $xfer_size = do headers ("mailer", 0);
      $xfer_size += do copy ("mailer");
      close mailer;
      sleep (30) unless $num == $count;	# allow system to stabilize
    }
  }
}

sub synchronise {

  # try to reduce system load by preventing this job to run
  # if there are other heavy tasks active

  return if $nosync;

  do {

    $notnow = 0;

    # sleep a while if -

    # - expire is running

    if ( opendir (xp, "/usr/spool/news/.rnews") ) {
      @xp = readdir (xp);
      closedir (xp);
      $notnow++ if $#xp > 1;
      printf stderr "[%05d] waiting for news spool to clean up ...\n", $$ if $debug;
    }

    # - uucp processes are running
    @uucp_locks = </usr/spool/uucp/LCK.*>;
    if (($#uucp_locks >= 0) && ($uucp_locks[0] ne "/usr/spool/uucp/LCK.*")) {
      $notnow++;
      printf stderr "[%05d] waiting for uucp tasks ...\n", $$ if $debug;
    }

    # - another process is running
    if ( $lockfile && ( $notnow == 0 )) {

      if ( -e $lockfile ) {

	# check if the process is still active
	open (lf, $lockfile);
	$pid = <lf>;
	close (lf);
	chop ($pid);
	if ( ( kill 0, $pid ) > 0 ) {	# it is ...
	  $notnow++;
	  printf stderr "[%05d] waiting for process $pid ...\n", $$ if $debug;
	}
	else {			# process disappeared - seize the lock
	  open (lf, ">" . $lockfile);
	  print lf "$$\n";
	  close (lf);
	  printf stderr "[%05d] seized lock of process $pid\n", $$ if $debug;
	}
      }
      else {
	# create a lock with our pid in it.
	open (lf, ">" . $lockfile);
	print lf "$$\n";
	close (lf);
      }
    }

    sleep ($debug ? 30 : 300) if $notnow;

  } while $notnow;
}

sub writelog {

  # Write message to logfile, if possible, Otherwise use stderr.

  @tm = localtime (time);
  $msg = sprintf ("%02d%02d%02d %02d:%02d %s\n", 
                  $tm[5], $tm[4]+1, $tm[3], $tm[2], $tm[1], $_[0]);

  if ( !$nolog && $logfile && ( -w $logfile ) && open (l, ">>" . $logfile) ) {
    print l $msg;
    close l;
    return unless $debug;
  }

  print stderr $msg;
}
