#!/usr/bin/perl -w

# @(#)@ scanindex	1.1 - scanindex.pl
#
# read INDEX file from stdin, and writes it , with size indication, to STDOUT.
# 
# also produces a consistency report.

do "stat.pl";
$S_IFREG = 0100000;	# regular file

while ( $line = <> ) {

  chop ($line);

  # skip empty lines and comment lines
  $ok = 1;
  next if $line =~ /^#/;
  next if length ($line) == 0;
  $ok = 0;

  # split and ignore lines which are not well formatted
  if ( $line =~ /(\t+ *)|(\t*  +)/ ) {
    $file = $`;
  } 
  else {
    next;
  }

  if ( ! do Stat ($file) ) {
    print STDERR "Could not stat: \"$file\"\n";
    next;
  }

  if ( ! ( ($st_mode & $S_IFREG) && ($st_size > 0) ) ) {
    print STDERR "Not a file or empty: \"$file\"\n";
    next;
  }

  # phew, we have something decent

  @file = split (/\//, $file);
  $file[$#file] = substr($file[$#file],0,14);
  $in_index{join('/',@file)} = 1;

  # strip existing size
  $line = $` if $line =~ / *\(\d+K\) *$/ ;
  # add new size
  $line .= " (" . int(($st_size+1023)/1024) . "K)";
  $ok = 1;

}
continue {
  print STDOUT ($ok ? "" : "# "), $line, "\n";
}

# now check for completeness

open (dir, "find * \\( -type f -o -type l \\) -print|");
@in_place = ();
while ( <dir> ) { chop; push (@in_place, $_); }
close (dir);
@not_found = ();
@ignored = ();

foreach $file ( @in_place ) {
  if ( $in_index{$file} ) {
    delete $in_index{$file};
  } else {
    if ( ( $file =~ /~$/ ) || ( $file =~ /#.+#$/ ) ) {
      push (@ignored, $file);
    }
    else {
      push (@not_found, $file);
    }
  }
}

#@index = sort keys (%in_index);
@x = keys (%in_index);
@index = sort @x;
@not_found = sort @not_found;

do printtab ("Entries in index but not on disk", @index);
do printtab ("Entries on disk but not in index", @not_found);
do printtab ("Ingored entries", @ignored);

sub printtab {
  local ($msg) = shift (@_);
  local ($file);
  return if  $#_ < 0;

  print STDERR "$msg\n\n";
  foreach $file ( @_ ) {
    print STDERR "\t$file\n";
  }
  print STDERR "\n";
}
